package de.uni_passau.fim.seibt.gitwrapper.repo;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.Optional;
import java.util.Scanner;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.ExecRes;
import org.apache.commons.io.FileUtils;

/**
 * A git {@link Repository}.
 */
public class Repository {

    private static final Logger LOG = Logger.getLogger(Repository.class.getCanonicalName());

    // Constants defining expected commit header fields when parsing the output of 'git blame --porcelain'.
    private static final String AUTHOR = "author";
    private static final String AUTHOR_MAIL = "author-mail";
    private static final String AUTHOR_TIME = "author-time";
    private static final String AUTHOR_TZ = "author-tz";
    private static final String COMMITTER = "committer";
    private static final String COMMITTER_MAIL = "committer-mail";
    private static final String COMMITTER_TIME = "committer-time";
    private static final String COMMITTER_TZ = "committer-tz";
    private static final String FILENAME = "filename";
    private static final String SUMMARY = "summary";
    private static final String BOUNDARY = "boundary";

    // The type returned for commits by 'git cat-file -t".
    private static final String TYPE_COMMIT = "commit";

    private static final String CONFLICT_START = "<<<<<<<";
    private static final String CONFLICT_MIDDLE = "=======";
    private static final String CONFLICT_END = ">>>>>>>";

    // The prefixes of the full symbolic names of (remote) branches.
    private static final String FULL_SYMBOLIC_BRANCH = "refs/heads/";
    private static final String FULL_SYMBOLIC_REMOTE_BRANCH = "refs/remotes/";

    private GitWrapper git;

    private String url;
    private File dir;

    private Map<String, Commit> commits;
    private Map<String, Branch> branches;

    /**
     * Constructs a new {@link Repository}.
     *
     * @param git
     *         the {@link GitWrapper} to use
     * @param url
     *         the url the {@link Repository} was cloned from
     * @param dir
     *         the directory containing the {@link Repository}
     */
    Repository(GitWrapper git, String url, File dir) {
        this.git = git;
        this.url = url;

        try {
            this.dir = dir.getCanonicalFile();
        } catch (IOException e) {
            LOG.log(Level.WARNING, e, () -> "Exception computing the canonical file for " + dir + ". Using the absolute file instead.");
            this.dir = dir.getAbsoluteFile();
        }

        this.commits = new HashMap<>();
        this.commits.put(DummyCommit.DUMMY_COMMIT_ID, new DummyCommit(this));
        this.branches = new HashMap<>();
    }

    /**
     * Performs a checkout of the given {@link Reference}.
     *
     * @param ref
     *         the {@link Reference} to checkout
     * @return whether the checkout was successful
     * @see <a href=https://git-scm.com/docs/git-checkout>git checkout</a>
     */
    public boolean checkout(Reference ref) {
        Optional<ExecRes> checkout = git.exec(dir, "checkout", ref.getId());
        Function<ExecRes, Boolean> toBoolean = res -> {
            boolean failed = git.failed(res);

            if (failed) {
                LOG.warning(() -> String.format("Checkout of %s failed.", ref));
            }

            return !failed;
        };

        return checkout.map(toBoolean).orElse(false);
    }

    /**
     * Performs a checkout of the given {@link Reference}. All changes since last commit will be discarded.
     *
     * @param ref
     *         the {@link Reference} to checkout
     * @return whether the checkout was successful
     * @see <a href=https://git-scm.com/docs/git-checkout>git checkout</a>
     */
    public boolean forceCheckout(Reference ref) {
        Optional<ExecRes> checkout = git.exec(dir, "reset", "--hard");
        Function<ExecRes, Boolean> toBoolean = res -> {
            boolean failed = git.failed(res);

            if (failed) {
                LOG.warning(() -> String.format("Reset of %s failed.", ref));
                return false;
            }

            return checkout(ref);
        };

        return checkout.map(toBoolean).orElse(false);
    }

    /**
     * Performs a fetch in this {@link Repository}.
     *
     * @return whether the fetch was successful
     * @see <a href=https://git-scm.com/docs/git-fetch>git fetch</a>
     */
    public boolean fetch() {
        Optional<ExecRes> fetch = git.exec(dir, "fetch", "--all", "--tags"); //TODO add "--prune"?
        Function<ExecRes, Boolean> toBoolean = res -> {
            boolean failed = git.failed(res);

            if (failed) {
                LOG.warning(() -> String.format("Fetch of %s failed.", this));
            }

            return !failed;
        };

        return fetch.map(toBoolean).orElse(false);
    }

    /**
     * Returns the merge commits in this repository.
     *
     * @return the merge commits
     */
    public List<Commit> getMergeCommits() {
        Optional<ExecRes> revList = git.exec(dir, "rev-list", "--all", "--merges");
        Function<ExecRes, List<Commit>> toCommitList = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to obtain the merge commits from %s.", this));
                return null;
            }

            String[] lines;

            if (res.getStdOutTrimmed().isEmpty()) {
                lines = new String[] {};
            } else {
                lines = res.getStdOutTrimmed().split("[\\r?\\n]+");
            }

            LOG.fine(() -> String.format("Found %d merge commits in %s.", lines.length, this));

            if (lines.length > 0) {
                LOG.finer(() -> String.format("Merge commits are:%n%s", String.join(System.lineSeparator(), lines)));
            }

            return Arrays.stream(lines).map(this::getCommitUnchecked).collect(Collectors.toList());
        };

        return revList.map(toCommitList).orElse(Collections.emptyList());
    }

    /**
     * Returns the {@link Commit} currently pointed at by <code>HEAD</code>.
     *
     * @return the <code>HEAD</code> commit
     */
    public Optional<Commit> getCurrentHEAD() {
        Optional<ExecRes> revParse = git.exec(dir, "rev-parse", "HEAD");
        Function<ExecRes, Commit> toHEAD = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> "Failed to obtain the current HEAD commit.");
                return null;
            }

            return getCommitUnchecked(res.getStdOutTrimmed());
        };

        return revParse.map(toHEAD);
    }

    /**
     * Returns a {@link Commit} for the given ID. The caller must ensure that the ID is a full SHA1 hash of a
     * commit that exists in this repository.
     *
     * @param id
     *         the ID for the {@link Commit}
     * @return the {@link Commit}
     */
    Commit getCommitUnchecked(String id) {
        return commits.computeIfAbsent(id, theID -> new Commit(this, theID));
    }

    /**
     * Optionally returns a {@link Commit} for the given ID. If the given ID does not designate a commit that exists in
     * this {@link Repository}, an empty {@link Optional} will be returned. The ID will be resolved to a full SHA1 hash.
     *
     * @param id
     *         the ID of the commit
     * @return the {@link Commit} or an empty {@link Optional} if the ID is invalid or an exception occurs
     */
    public Optional<Commit> getCommit(String id) {
        if (commits.containsKey(id)) {
            return Optional.of(commits.get(id));
        }

        return verify(false, id, TYPE_COMMIT).map(sha1 -> commits.computeIfAbsent(sha1, fullID -> new Commit(this, fullID)));
    }

    /**
     * Optionally returns a {@link Branch} for the given name. If the given name does not designate a branch that
     * exists in this {@link Repository}, an empty {@link Optional} will be returned.
     *
     * @param name
     *         the name of the branch
     * @return the {@link Branch} or an empty {@link Optional} if the name is invalid or an exception occurs
     */
    public Optional<Branch> getBranch(String name) {
        if (branches.containsKey(name)) {
            return Optional.of(branches.get(name));
        }

        Function<String, Branch> toBranch = fullName -> {
            if (fullName.isEmpty()) {
                LOG.warning(() -> String.format("The name '%s' does not designate a branch in %s.", name, this));
                return null;
            }

            String branchName;

            if (fullName.startsWith(FULL_SYMBOLIC_BRANCH)) {
                branchName = fullName.substring(FULL_SYMBOLIC_BRANCH.length());
            } else if (fullName.startsWith(FULL_SYMBOLIC_REMOTE_BRANCH)) {
                branchName = fullName.substring(FULL_SYMBOLIC_REMOTE_BRANCH.length());
            } else {
                LOG.warning(() -> String.format("The name '%s' designates neither a local nor a remote branch in %s.", fullName, this));
                return null;
            }

            return branches.computeIfAbsent(branchName, newBranchName -> new Branch(this, newBranchName));
        };

        return verify(true, name, null).map(toBranch);
    }

    /**
     * Resolves the given <code>id</code> to the full SHA1 hash.
     *
     * @param id
     *         the <code>id</code> to transform
     * @return the full SHA1 hash or an empty {@link Optional} if an exception occurs or {@code id} can not be converted
     *         to its associated git hash
     */
    Optional<String> toHash(String id) {
        return verify(false, id, null);
    }

    /**
     * Verifies that the given git-thing exists and can be resolved to a raw SHA1 hash. If type is not {@link null},
     * git will be asked to verify that the thing is also of the given type. If {@code fullSymbolicName} is set, the
     * returned {@code String} will be (if present) the full symbolic name of the reference. If {@code arg} does
     * not designate a reference but does exist in the repository, an empty {@code String} will be returned. This
     * is the case (for example) for an existing commit in the repository.
     *
     * @param fullSymbolicName
     *         whether to return the full symbolic name of {@code arg} if possible
     * @param arg
     *         the thing to check
     * @param type
     *         the type of the thing, ignored if {@code null}
     * @return optionally the full SHA1 hash (or full symbolic name) or an empty {@link Optional} if the thing can not
     *         be found or turned into a full hash
     */
    private Optional<String> verify(boolean fullSymbolicName, String arg, String type) {
        String argument = type != null ? String.format("%s^{%s}", arg, type) : arg;
        Optional<ExecRes> revParse;

        if (fullSymbolicName) {
            revParse = git.exec(dir, "rev-parse", "--symbolic-full-name", "--verify", argument);
        } else {
            revParse = git.exec(dir, "rev-parse", "--verify", argument);
        }

        Function<ExecRes, String> toHash = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to resolve '%s' to its unique SHA1 hash.", arg));
                return null;
            }

            String hash = res.getStdOutTrimmed();
            LOG.finer(() -> String.format("Verified that %s exists and resolved it to %s.", arg, hash));

            return hash;
        };

        return revParse.map(toHash);
    }

    /**
     * Copies this {@link Repository} to the given <code>destination</code> directory.
     *
     * @param destination
     *         the destination directory
     * @return a copy of this {@link Repository} at the new location
     * @see FileUtils#copyDirectory(File, File)
     */
    public Optional<Repository> copy(File destination) {
        try {
            if (destination.exists() && destination.isDirectory()) {
                LOG.warning(() -> String.format("%s already exists. Merging source and destination directories.", destination));
            }

            FileUtils.copyDirectory(dir, destination);

            LOG.fine(() -> String.format("Copied %s to %s.", this, destination));
        } catch (IOException e) {
            LOG.log(Level.WARNING, e, () -> String.format("Failed to copy %s to %s.", dir, destination));
            return Optional.empty();
        }

        Repository repo = new Repository(git, url, destination);

        for (Map.Entry<String, Commit> entry : commits.entrySet()) {
            repo.getCommitUnchecked(entry.getValue().getId());
        }

        return Optional.of(repo);
    }

    /**
     * Parses the output of 'git blame' for the given {@code file} and returns a list of {@link BlameLine BlameLines}
     * containing the resulting information about every line of the file. If there is an exception parsing or obtaining
     * the 'git blame' output, an empty {@link Optional} will be returned.
     *
     * @param file
     *         the path to the file to analyze (relative to the git root or absolute)
     * @return optionally the list of {@link BlameLine BlameLines}
     */
    public Optional<List<BlameLine>> blameFile(Path file) {
        Path path = dir.toPath().relativize(dir.toPath().resolve(file));
        Optional<ExecRes> output = git.exec(dir, "blame", "-p", path.toString());
        Function<ExecRes, List<BlameLine>> toList = execRes -> {

            if (git.failed(execRes)) {
                LOG.warning(() -> String.format("Failed to get blame information for file %s", path));
                return null;
            }

            List<BlameLine> bLines = null;

            try {
                bLines = parseBlameLines(execRes.stdOut);
            } catch (NoSuchElementException e) {
                LOG.log(Level.WARNING, e, () -> String.format("Failed to parse the blame information for file %s", path));
            }

            return bLines;
        };

        return output.map(toList);
    }

    /**
     * Parses the given 'git blame' output and returns a list of {@link BlameLine BlameLines} containing the resulting
     * information about every line of the output.
     *
     * @param blameOutput
     *         the 'git blame' output to parse
     * @return the list of {@link BlameLine BlameLines}
     */
    private List<BlameLine> parseBlameLines(String blameOutput) {
        Scanner lines = new Scanner(blameOutput);

        if (!lines.hasNextLine()) {
            return Collections.emptyList();
        }

        List<BlameLine> bLines = new ArrayList<>();
        Map<Commit, BlameLine> firstOccurrences = new HashMap<>();

        while (lines.hasNextLine()) {
            String line;
            Scanner lineScanner;

            // Read the first line of the commit header and construct/get the Commit object.
            line = lines.nextLine();
            lineScanner = new Scanner(line);

            String sha1 = lineScanner.next();
            int originalLineNumber = lineScanner.nextInt();
            int finalLineNumber = lineScanner.nextInt();

            Commit commit = getCommitUnchecked(sha1);

            Instant authorInstant = null;
            Instant committerInstant = null;
            ZoneId authorTZ = null;
            ZoneId committerTZ = null;

            File commitFile = null;

            Map<String, String> other = new HashMap<>();

            // Read the rest of the commit header and input all fields into the Commit object or the BlameLine fields.
            while (!(line = lines.nextLine()).startsWith("\t")) {
                lineScanner = new Scanner(line);
                String headerKey = lineScanner.next();

                switch (headerKey) {
                    case AUTHOR:
                        commit.setAuthor(lineScanner.nextLine().trim());
                        break;
                    case AUTHOR_MAIL: {
                        String email = lineScanner.nextLine().trim();
                        commit.setAuthorMail(email.substring(1, email.length() - 1));
                    } break;
                    case AUTHOR_TIME:
                        authorInstant = Instant.ofEpochSecond(lineScanner.nextLong());
                        break;
                    case AUTHOR_TZ:
                        authorTZ = ZoneId.of(lineScanner.next());
                        break;
                    case COMMITTER:
                        commit.setCommitter(lineScanner.nextLine().trim());
                        break;
                    case COMMITTER_MAIL: {
                        String email = lineScanner.nextLine().trim();
                        commit.setCommitterMail(email.substring(1, email.length() - 1));
                    } break;
                    case COMMITTER_TIME:
                        committerInstant = Instant.ofEpochSecond(lineScanner.nextLong());
                        break;
                    case COMMITTER_TZ:
                        committerTZ = ZoneId.of(lineScanner.next());
                        break;
                    case FILENAME:
                        commitFile = dir.toPath().resolve(Paths.get(lineScanner.nextLine().trim())).toFile();
                        break;
                    case SUMMARY:
                        // Skip the summary of the commit as it is just the first line of its message (for which there is a getter).
                        break;
                    case BOUNDARY:
                        // We do not store the information that this commit was the boundary of the 'git blame' call.
                        break;
                    default:
                        other.put(headerKey, lineScanner.hasNextLine() ? lineScanner.nextLine().trim() : "");
                }
            }

            if (authorInstant != null && authorTZ != null) {
                commit.setAuthorTime(OffsetDateTime.ofInstant(authorInstant, authorTZ));
            }

            if (committerInstant != null && committerTZ != null) {
                commit.setCommitterTime(OffsetDateTime.ofInstant(committerInstant, committerTZ));
            }

            // Read the line starting with a TAB and store it in a BlameLine along with its commit.
            BlameLine bLine;
            line = line.substring(1); // Remove the TAB.

            if (firstOccurrences.containsKey(commit)) {
                bLine = BlameLine.nextOccurrence(firstOccurrences.get(commit), originalLineNumber, finalLineNumber, line);
            } else {
                bLine = BlameLine.firstOccurrence(commit, originalLineNumber, finalLineNumber, commitFile, line, other);
                firstOccurrences.put(commit, bLine);
            }

            bLines.add(bLine);
        }

        return bLines;
    }

    /**
     * Parses the output of 'git blame' for an unmerged file and extracts the merge conflicts that occurred in the file.
     * Every merge conflict is represented by a {@link MergeConflict} instance containing the {@link BlameLine BlameLines}
     * that represent the lines involved in the conflict. If there are no conflicts in the given file, an empty list
     * will be returned. If there is an exception parsing or obtaining the 'git blame' output, an empty {@link Optional}
     * will be returned.
     *
     * @param file
     *         the path to the unmerged file to analyze (relative to the git root or absolute)
     * @return optionally the list of {@link MergeConflict MergeConflicts}
     * @see #blameFile(Path)
     */
    public Optional<List<MergeConflict>> blameUnmergedFile(Path file) {
        Optional<List<BlameLine>> blame = blameFile(file);

        if (!blame.isPresent()) {
            return Optional.empty();
        }

        List<MergeConflict> conflicts = new ArrayList<>();

        List<BlameLine> left = null;
        List<BlameLine> right = null;
        String lastMarker = null;

        for (BlameLine line : blame.get()) {
            if (line.line.startsWith(CONFLICT_START)) {
                left = new ArrayList<>();
                right = new ArrayList<>();
                lastMarker = CONFLICT_START;
            } else if (line.line.startsWith(CONFLICT_MIDDLE)) {
                lastMarker = CONFLICT_MIDDLE;
            } else if (line.line.startsWith(CONFLICT_END)) {
                conflicts.add(new MergeConflict(left, right));
                lastMarker = CONFLICT_END;
            } else if (CONFLICT_START.equals(lastMarker)) {
                assert left != null; left.add(line);
            } else if (CONFLICT_MIDDLE.equals(lastMarker)) {
                assert right != null; right.add(line);
            }
        }

        return Optional.of(conflicts);
    }

    /**
     * Parses and returns the current output of 'git status' for this {@link Repository}.
     * If there is an exception parsing or obtaining the status output, an empty {@link Optional} will be returned.
     *
     * @return optionally the {@link Status} of this {@link Repository}
     */
    public Optional<Status> getStatus() {
        Optional<ExecRes> output = git.exec(dir, "status", "-z");
        Function<ExecRes, Status> toStatus = execRes -> {

            if (git.failed(execRes)) {
                LOG.warning(() -> String.format("Failed to get status information for repo %s", this));
                return null;
            }

            return Status.parseStatus(this, execRes.stdOut+"\0"); //TODO Parser so anpassen, dass kein \0 erforderlich ist.
        };

        return output.map(toStatus);
    }

    /**
     * Returns the {@link GitWrapper} used by this {@link Repository}.
     *
     * @return the {@link GitWrapper}
     */
    GitWrapper getGit() {
        return git;
    }

    /**
     * Returns the URL this {@link Repository} was cloned from.
     *
     * @return the URL
     */
    public String getUrl() {
        return url;
    }

    /**
     * Returns the directory containing this {@link Repository}.
     *
     * @return the repository directory
     */
    public File getDir() {
        return dir;
    }

    /**
     * Returns the name of the directory this {@link Repository} is based in.
     *
     * @return the name of the {@link Repository}
     */
    public String getName() {
        return getDir().getName();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        Repository that = (Repository) o;
        return Objects.equals(url, that.url) && Objects.equals(dir, that.dir);
    }

    @Override
    public int hashCode() {
        return Objects.hash(url, dir);
    }

    @Override
    public String toString() {
        return String.format("Repository{url=%s, dir=%s}", url, dir);
    }
}
