package de.uni_passau.fim.seibt.gitwrapper.repo;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class Status {
    private static final Pattern STATUS_ENTRY = Pattern.compile("(.?.) (?:.*? )?(.*?)\0");

    public final String id;
    public final Map<Path, String> changed;
    public final Map<Path, String> unmerged;

    /**
     * @param id       the commit id
     * @param changed  a map of changed files and heir status codes
     * @param unmerged a map of unmerged files an their status code.
     */
    private Status(String id, Map<Path, String> changed, Map<Path, String> unmerged) {
        this.id = id;
        this.changed = Collections.unmodifiableMap(changed);
        this.unmerged = Collections.unmodifiableMap(unmerged);
    }

    /**
     * Checks, if <code>this</code> represents a clean repository.
     *
     * @return <code>true</code>, if the repository state is clean and has uncommitted changes.
     */
    public boolean isClean() {
        return changed.isEmpty() && unmerged.isEmpty();
    }

    /**
     * Checks, if <code>this</code> has unmerged files.
     *
     * @return <code>true</code>, if there are unmerged files
     */
    public boolean hasConflicts() {
        return !unmerged.isEmpty();
    }

    @Override
    public String toString() {
        String out = "";
        if (isClean()) {
            out = "Clean working directory on commit " + id + "\n";
        } else {
            if (!unmerged.isEmpty()) {
                out = "Unmerged files:";
                out += unmerged.keySet().stream().map(Path::toString).reduce("", (list, file) -> String.join("\n", list, file));
                out += "\n";
            }
            if (!changed.isEmpty()) {
                out += "Changed files:";
                out += changed.keySet().stream().map(Path::toString).reduce("", (list, file) -> String.join("\n", list, file));
                out += "\n";
            }
        }

        return out;
    }

    /**
     * Parses the output from git into a new object.
     *
     * @param repo      the repo this status report is from
     * @param gitOutput the status report from git
     * @return a {@link Status} object, representing the status read from the git output
     */
    static Status parseStatus(Repository repo, String gitOutput) {
        Map<Path, String> changed = new HashMap<>();
        Map<Path, String> unmerged = new HashMap<>();
        Matcher matcher = STATUS_ENTRY.matcher(gitOutput);
        while (matcher.find()) {
            // used new path, if file was moved
            Path file = Paths.get(matcher.group(2));
            String code = matcher.group(1).toUpperCase();
            if (code.contains("U")) {
                unmerged.put(file, code);
            } else {
                changed.put(file, code);
            }
        }

        return new Status(repo.getCurrentHEAD().get().getId(), changed, unmerged);
    }
}
